#include "stdafx.h"
#include "directory.h"
#include <vector>
#include <algorithm>
#include "undeclared.h"
#include "messages.h"

#define g_objectDirPrefix L"\\BaseNamedObjects\\ChrootDir"

static HANDLE PointDriveAtDir(HANDLE hRootDir, PUNICODE_STRING pDir, DWORD index)
{
	ASSERT(index < 26);
	WCHAR driveRoot[] = {static_cast<WCHAR>(L'A' + index), L':'};
	UNICODE_STRING driveString = {sizeof(driveRoot), sizeof(driveRoot), driveRoot};
	OBJECT_ATTRIBUTES oa;
	// we set OBJ_INHERIT here so the handle count of these symlinks will increase
	// when (or indeed, if) we create the child process. We can then close
	// our bunch of handles and let the lifetime of the child
	// govern the lifetime of the symlinks. It also means we don't have
	// to stick around doing a WaitForSingleObject on the launched process
	// just to cleanup when it exits.
	InitializeObjectAttributes(&oa, &driveString, OBJ_CASE_INSENSITIVE | OBJ_INHERIT, hRootDir, NULL);
	HANDLE hLink = NULL;
	NTSTATUS stat = NtCreateSymbolicLinkObject(&hLink, 0, &oa, pDir);
	BOOL ret = NT_SUCCESS(stat);
	if(!ret)
	{
		// note to Microsoft, this doesn't actually set the
		// the last NTSTATUS value, just the last win32 error
		// for everybody else, yep there's a field for each in the TEB
		RtlSetLastWin32ErrorAndNtStatusFromNtStatus(stat);
	}
	return hLink;
}

static BOOL PointDrivesAtDir(HANDLE hRootDir, DWORD drivesToKeep, PUNICODE_STRING pDir, std::vector<HANDLE>& createdSymlinks)
{
	// ensure we don't get fed with a dos path
	// nothing can find anything when that happens
	ASSERT(wmemchr(pDir->Buffer, L':', pDir->Length / sizeof(WCHAR)) == NULL);
	DWORD driveBitmap = GetLogicalDrives();
	DWORD index = 0;
	while(BitScanForward(&index, driveBitmap))
	{
		ASSERT(index < 26);
		if(!BitTest(reinterpret_cast<LONG*>(&drivesToKeep), index))
		{
			HANDLE hLink = PointDriveAtDir(hRootDir, pDir, index);
			if(!hLink)
			{
				ERROR_TEXT("Failed to redirect drive %c: - %lu", 'A' + index, GetLastError());
				return FALSE;
			}
			createdSymlinks.push_back(hLink);
		}
		driveBitmap ^= (1 << index);
	}
	return TRUE;
}

static void TranslateDosPathToDevicePath(std::wstring& dosPath)
{
	ASSERT(!dosPath.empty());
	ASSERT(RtlDetermineDosPathNameType_U(dosPath.c_str()) == RtlPathTypeDriveAbsolute);
	WCHAR driveLetter[] = {dosPath[0], dosPath[1], 0};
	WCHAR devicePath[MAX_PATH] = {0};
	QueryDosDevice(driveLetter, devicePath, ARRAYSIZE(devicePath));
	// replace the drive letter and the colon with the device path
	dosPath.replace(0, 2, devicePath);
}

HANDLE CreateObjectDirectory(std::wstring& jailDir, DWORD drivesToKeep, std::vector<HANDLE>& driveSymlinks)
{
	WCHAR dirName[ARRAYSIZE(g_objectDirPrefix) + std::numeric_limits<DWORD>::digits10 + 1];
	size_t dirNameChars = ARRAYSIZE(dirName);
	size_t remaining = 0;
	DWORD currentPid = GetCurrentProcessId();
	HRESULT hr = StringCchPrintfExW(dirName, dirNameChars, NULL, &remaining, 0, g_objectDirPrefix L"%lu", currentPid);
	ASSERT(SUCCEEDED(hr));

	USHORT stringBytes = static_cast<USHORT>((dirNameChars - remaining) * sizeof(WCHAR));
	UNICODE_STRING objectDir = {stringBytes, stringBytes, dirName};
	OBJECT_ATTRIBUTES oa;
	InitializeObjectAttributes(&oa, &objectDir, OBJ_CASE_INSENSITIVE | OBJ_INHERIT, NULL, NULL);
	WinType<> hDir(NULL, &CloseHandle);
	NTSTATUS stat = NtCreateDirectoryObject(&hDir, DIRECTORY_ALL_ACCESS, &oa);
	if(!NT_SUCCESS(stat))
	{
		DisplayError("Failed to create object directory", stat);
		return NULL;
	}

	TranslateDosPathToDevicePath(jailDir);

	// don't need to nul a counted unicode_string
	stringBytes = static_cast<USHORT>(jailDir.length() * sizeof(WCHAR));
	UNICODE_STRING jailString = {stringBytes, stringBytes, &jailDir[0]};
	if(!PointDrivesAtDir(*hDir, drivesToKeep, &jailString, driveSymlinks))
	{
		DisplayError("Failed to create object directory", stat);
		std::for_each(driveSymlinks.begin(), driveSymlinks.end(), &CloseHandle);
		driveSymlinks.clear();
		return NULL;
	}
	return hDir.Release();
}

void DuplicateHandleToChildAndClose(HANDLE hTargetProcess, HANDLE hHandle)
{
	HANDLE hCurProc = GetCurrentProcess();
	HANDLE hTarget = NULL;
	// The handle is set to inherit so any child processes
	// will keep the symlinks extant until they exit
	if(!DuplicateHandle(hCurProc, hHandle, hTargetProcess, &hTarget, READ_CONTROL, TRUE, DUPLICATE_CLOSE_SOURCE))
	{
		DBG_TEXT("Couldn't copy symlink handle to child - %lu", GetLastError());
	}
}
