#include "stdafx.h"
#include <cctype>
#include <algorithm>
#include <functional>
#include <vector>
#include <sstream>
#include "user.h"
#include "unpack.h"
#include "directory.h"
#include "undeclared.h"
#include "messages.h"

static const LPCWSTR CHROOT_VERSION_BANNER = L"Chroot v1.00 - http://www.airesoft.co.uk";
static LPWSTR CHROOT_DEFAULT_COMMAND = L"%ComSpec%";

static BOOL HandleCommandLine(wchar_t** argv, CmdLineArgs& args)
{
	// argv is at least two elements long
	// due to the check in main
	LPCWSTR pPotentialDir = *argv;
	wchar_t** ppFirstArgNotSeen = argv + 1;
	while(*argv)
	{
		LPCWSTR pArg = *argv;
		if((*pArg == L'-') || (*pArg == L'/'))
		{
			++pArg;
			bool reexamine;
			do
			{
				reexamine = false;
				switch(towlower(*pArg))
				{
					case L'-':
					case L'/':
					{
						++pArg;
						reexamine = true;
					}
					break;
					case L'v':
					{
						_putws(CHROOT_VERSION_BANNER);
						exit(0);
					}
					break;
					case L'h':
					case L'?':
					{
						return FALSE;
					}
					break;
					case L'a':
					{
						args.createUser = true;
						if(args.promptForUser)
						{
							return FALSE;
						}
					}
					break;
					case L'u':
					{
						args.promptForUser = true;
						LPCWSTR pUserName = (*(++argv));
						if(args.createUser || pUserName == NULL)
						{
							return FALSE;
						}
						if(!DoesUserExist(pUserName))
						{
							DisplayLastError("Username lookup failed");
							exit(1);
						}
						args.userName = pUserName;
					}
					break;
					case L'f':
					{
						DWORD driveMask = 0;
						LPCWSTR pDrives = (*(++argv));
						WCHAR drive;
						while((drive = *pDrives) != 0)
						{
							// daft 'apeths
							// if they can't read the docs
							// they don't deserve to succeed
							if(!iswalpha(drive))
							{
								return FALSE;
							}
							drive = towlower(drive);
							WCHAR index = drive - L'a';
							driveMask |= (1 << index);
							++pDrives;
						}
						args.drivesToKeep = driveMask;
					}
					break;
				}
			}
			while(reexamine);
			ppFirstArgNotSeen = argv + 1;
		}
		++argv;
	}
	args.dir = ExpandEnvString(pPotentialDir);
	// no command given, use cmd as the default
	if(ppFirstArgNotSeen == argv)
	{
		args.ppProgArgs = &CHROOT_DEFAULT_COMMAND;
		args.numProcArgs = 1;
	}
	// otherwise use the unprocessed command line as the program and args
	else
	{
		args.ppProgArgs = ppFirstArgNotSeen;
		args.numProcArgs = static_cast<int>(argv - ppFirstArgNotSeen);
	}
	return TRUE;
}

static BOOL RunCommand(HANDLE hDir, CmdLineArgs& args, std::vector<HANDLE>& symlinks)
{
	std::wostringstream cmdParams;
	std::wstring programCmdLine = ExpandEnvString(args.ppProgArgs[0]);
	programCmdLine.push_back(0); // nul terminate 
	args.ppProgArgs[0] = &programCmdLine[0]; // so that this is fine
	for(int i = 0; i < args.numProcArgs; ++i)
	{
		// quote all arguments just in case
		cmdParams << L'\"' << args.ppProgArgs[i] << L"\" ";
	}
	// ensure the string is null terminated, since we'll be using
	// &str[0] instead of c_str(), there's no guarantee the contents will be terminated
	cmdParams << L'\0';
	programCmdLine = cmdParams.str();
	PROCESS_INFORMATION pi = {0};
	BOOL createProcRet = FALSE;
	// not a typo
	if(args.createUser | args.promptForUser)
	{
		createProcRet = RunProcessAsUser(programCmdLine, args, pi);
		if(createProcRet)
		{
			// the 'CreateProcessAsXxx' and 'CreateProcessWithXxx' don't allow
			// the inheriting of handles (because they're launched using a service)
			// so we copy the handles to the child process here
			std::for_each(
				symlinks.begin(),
				symlinks.end(),
				std::bind1st(
					std::ptr_fun(&DuplicateHandleToChildAndClose),
					pi.hProcess
				)
			);
		}
	}
	else
	{
		WCHAR rootDir[4] = {FindARootedDriveLetter(args.drivesToKeep), L':', L'\\', 0};
		STARTUPINFO si = {sizeof(si), 0};
		si.dwFlags = STARTF_USESHOWWINDOW;
		si.wShowWindow = SW_SHOW;
		createProcRet = CreateProcess(
			NULL,
			&programCmdLine[0],
			NULL,
			NULL,
			TRUE,
			CREATE_SUSPENDED | CREATE_NEW_CONSOLE | CREATE_DEFAULT_ERROR_MODE | CREATE_NEW_PROCESS_GROUP,
			NULL,
			rootDir,
			&si,
			&pi
		);
		// now the handles have been inherited (or it failed) we no
		// longer need them
		std::for_each(symlinks.begin(), symlinks.end(), &CloseHandle);
	}
	if(!createProcRet)
	{
		DisplayLastError("Couldn't launch process");
		return FALSE;
	}
	WinType<> thread(pi.hThread, &CloseHandle);
	WinType<> process(pi.hProcess, &CloseHandle);
	NTSTATUS stat = NtSetInformationProcess(*process, ProcessDeviceMap, &hDir, sizeof(hDir));
	if(!NT_SUCCESS(stat))
	{
		TerminateProcess(*process, static_cast<UINT>(stat));
		DisplayError("Couldn't chroot new process", stat);
		return FALSE;
	}
	ResumeThread(*thread);
	DisplayStatus("Process running...");
	if(args.createUser)
	{
		// you can delete a user while it has a process running
		DeleteUser(args.userName);
	}
	return TRUE;
}

static BOOL CheckIsNativeApp()
{
#ifdef _WIN64
	return TRUE; // 64-bit app is always native at this point in history
#else
	// could use IsWow64Process but that's not available further downlevel than XP SP2
	SYSTEM_INFO sysInf = {0};
	GetNativeSystemInfo(&sysInf);
	return sysInf.wProcessorArchitecture == PROCESSOR_ARCHITECTURE_INTEL;
#endif
}

int __cdecl wmain(int argc, wchar_t** argv)
{
	CmdLineArgs args = {0};
	if(!(argc >= 2 && HandleCommandLine(argv + 1, args)))
	{
		PrintHelp();
		return 1;
	}
	if(!CheckIsNativeApp())
	{
		return Run64BitVersion(argc, argv);
	}
	if(!InitNtDllFnPtrs())
	{
		DisplayError("Couldn't find ntdll function pointers!");
		return 1;
	}
	if(args.createUser && !CheckCreateUserPrivilege())
	{
		DisplayError("Can't run as an anonymous user - you don't have the privilege to create a user!");
		return 1;
	}
	if(!GetUsernameAndPassword(args))
	{
		DisplayLastError("Unable to create user/get password");
		return 1;
	}
	std::vector<HANDLE> driveSymlinks;
	WinType<> hNewDir(CreateObjectDirectory(args.dir, args.drivesToKeep, driveSymlinks), &CloseHandle);
	if(!*hNewDir)
	{
		return 1;
	}
	if(!RunCommand(*hNewDir, args, driveSymlinks))
	{
		return 1;
	}
	return 0;
}