#ifndef CHROOT_SCOPED_OBJECTS_H
#define CHROOT_SCOPED_OBJECTS_H

#pragma once

template<class Type>
class ScopedArray
{
private:
	Type* data;

public:
	ScopedArray(Type* data) : data(data)
	{}

	ScopedArray(UINT elements) : data(new Type[elements])
	{}

	ScopedArray() : data(NULL)
	{}

	~ScopedArray()
	{		
		delete [] data;
	}

	Type* get() const
	{		
		ASSERT(data && "Array not initialized");
		return data;
	}

	Type** operator&()
	{		
		ASSERT(!data && "Array already initialized");
		return &data;
	}

	Type* operator*() const
	{		
		return get();
	}

	void Init(Type* newData)
	{		
		ASSERT(!data && "Array already initialized");
		data = newData;
	}

	Type* Release()
	{
		Type* pTemp = data;
		data = NULL;
		return pTemp;
	}

	void Swap(ScopedArray& other)
	{
		Type* pTemp = other.data;
		other.data = data;
		data = pTemp;
	}
};

template<class Type = HANDLE, class Ret = BOOL>
class WinType
{
private:
	typedef Ret (WINAPI*ReleaseProc)(Type);
	typedef void (WinType:: * bool_type)(WinType&);

	Type data;
	ReleaseProc lpfnReleaser;

	template<class A, class B>
	struct is_same
	{
		static const bool value = false;
	};

	template<class A>
	struct is_same<A, A>
	{
		static const bool value = true;
	};

public:
	WinType(const Type& data, ReleaseProc lpfnReleaser)
		: data(data),
		lpfnReleaser(lpfnReleaser)
	{}

	WinType(ReleaseProc lpfnReleaser)
		: data(),
		lpfnReleaser(lpfnReleaser)
	{}

	WinType()
		: data(),
		lpfnReleaser(NULL)
	{}

	~WinType()
	{
		// do nothing if the type is a handle, 
		// and it's NULL, closing that will lead to debugger breaks
		if(is_same<Type, HANDLE>::value && !data)
		{
			return;
		}
		if(lpfnReleaser)
		{
			lpfnReleaser(data);
		}
	}

	Type get() const
	{
		ASSERT(lpfnReleaser && "Object not initialised");
		return data;
	}

	Type operator*() const
	{		
		return get();
	}

	Type* operator&()
	{		
		ASSERT((data == Type()) && lpfnReleaser && "Object already initialised");
		return &data;
	}

	operator bool_type() const
	{
		return lpfnReleaser ? &WinType::Swap : NULL;
	}

	void Init(const Type& newData, ReleaseProc lpfnNewReleaser)
	{
		ASSERT((!lpfnReleaser) && "Object already initialized");
		data = newData;
		lpfnReleaser = lpfnNewReleaser;
	}

	Type Release()
	{		
		lpfnReleaser = NULL;
		Type temp = data;
		data = Type();
		return temp;
	}

	void Swap(WinType& other)
	{
		Type temp = other.data;
		ReleaseProc tempProc = other.lpfnReleaser;
		other.data = data;
		other.lpfnReleaser = lpfnReleaser;
		data = temp;
		lpfnReleaser = tempProc;
	}
};

#endif
