#include "stdafx.h"
#include "user.h"
#include <cstdlib>
#include <algorithm>
#include <ntsecapi.h>
#include <userenv.h>
#include <wincred.h>
#include <lm.h>
#include "messages.h"

// picks a character from the 62 available
static WCHAR GetRandomChar()
{
    static WCHAR alphabet[] = L"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789";
    int index = rand() % (ARRAYSIZE(alphabet) - 1);
    return alphabet[index];
}

static void GenerateUser(CmdLineArgs& args)
{
	PBYTE pBuffer = NULL;
	DWORD err = NetUserModalsGet(NULL, 0, &pBuffer);
	DWORD passwordLen = 8;
	if(err != NERR_Success)
	{
		DBG_TEXT("Getting user modals failed, using 8 as password length - %lu", err);
	}
	else
	{
		USER_MODALS_INFO_0* pModals = reinterpret_cast<USER_MODALS_INFO_0*>(pBuffer);
		passwordLen = std::max(pModals->usrmod0_min_passwd_len, passwordLen);
		NetApiBufferFree(pBuffer);
	}
	srand(GetTickCount());
	static const DWORD USER_NAME_LEN = 4;
    // simply generates two random strings
    // one of 4 letters in length (the user name) and
    // one of 8 letters in length (the password)
    std::wstring& user = args.userName;
    user.resize(USER_NAME_LEN);
    std::generate_n(&user[0], USER_NAME_LEN, &GetRandomChar);
	std::wstring& pass = args.password;
	pass.resize(passwordLen);
    std::generate_n(&pass[0], passwordLen, &GetRandomChar);
	USER_INFO_1 usrInf = {0};
	// try and create a simple user account on the target computer
    usrInf.usri1_name = const_cast<LPWSTR>(user.c_str());
    usrInf.usri1_password = const_cast<LPWSTR>(pass.c_str());
    usrInf.usri1_priv = USER_PRIV_USER;
    usrInf.usri1_flags = UF_PASSWD_CANT_CHANGE | UF_NORMAL_ACCOUNT;
	DWORD errParam = 0;
    err = NetUserAdd(NULL, 1, reinterpret_cast<PBYTE>(&usrInf), &errParam);
	if(err != NERR_Success)
	{
		DBG_TEXT("NetUserAdd failed on param %lu with error %lu", err, errParam);
		user.resize(0);
		pass.resize(0);
		return;
	}
	RtlEncryptMemory(&pass[0], static_cast<ULONG>(pass.length() * sizeof(WCHAR)), 0);
}

BOOL GetUsernameAndPassword(CmdLineArgs& args)
{
	if(args.createUser)
	{
		GenerateUser(args);
		return !(args.userName.empty() && args.password.empty());
	}
	else if(args.promptForUser)
	{
		DWORD passLen = CREDUI_MAX_PASSWORD_LENGTH + 1;
		std::wstring& userName = args.userName;
		// despite what it may look like, the max lengths
		// don't include the terminating null
		userName.resize(CREDUI_MAX_USERNAME_LENGTH + 1);
		DWORD nameLen = static_cast<DWORD>(userName.length());
		std::wstring& password = args.password;
		password.resize(passLen);
		DWORD flags = 
			CREDUI_FLAGS_DO_NOT_PERSIST | 
			CREDUI_FLAGS_EXCLUDE_CERTIFICATES |
			CREDUI_FLAGS_USERNAME_TARGET_CREDENTIALS | 
			CREDUI_FLAGS_INCORRECT_PASSWORD |
			CREDUI_FLAGS_KEEP_USERNAME;
		DWORD err = CredUICmdLinePromptForCredentials(
			args.ppProgArgs[0],
			NULL,
			0,
			&userName[0],
			nameLen,
			&password[0],
			passLen,
			NULL,
			flags
		);
		if(err)
		{
			DBG_TEXT("Failed to get password, err = %lu", err);
			return FALSE;
		}
		DWORD encryptMemoryWchars = RTL_ENCRYPT_MEMORY_SIZE / sizeof(WCHAR);
		// find the end of the password, as delimited by the nul
		std::wstring::size_type endOfPass = password.find(static_cast<WCHAR>(0));
		ASSERT(endOfPass != std::wstring::npos);
		// round the buffer to a multiple of RTL_ENCRYPT_MEMORY_SIZE bytes
		endOfPass = (endOfPass + encryptMemoryWchars) & ~(encryptMemoryWchars - 1);
		password.resize(endOfPass);
		// and encrypt the password
		RtlEncryptMemory(&password[0], static_cast<DWORD>(endOfPass) * sizeof(WCHAR), 0);
	}
	return TRUE;
}

BOOL CheckCreateUserPrivilege()
{
	// check if the user is an administrator or a member of the account operator group
	BYTE adminSidBuffer[SECURITY_MAX_SID_SIZE], acctOpsSidBuffer[SECURITY_MAX_SID_SIZE];
	PSID pAdminSid = adminSidBuffer, pAcctOpsSid = acctOpsSidBuffer;
	DWORD adminSidSize = sizeof(adminSidBuffer), acctOpsSidSize = sizeof(acctOpsSidBuffer);
	
	if(!(CreateWellKnownSid(WinBuiltinAccountOperatorsSid, NULL, pAcctOpsSid, &acctOpsSidSize) && 
		 CreateWellKnownSid(WinBuiltinAdministratorsSid, NULL, pAdminSid, &adminSidSize))
	)
	{
		DBG_TEXT("Couldn't create well known sids - %lu", GetLastError());
		return FALSE;
	}
	HANDLE hToken = NULL;
	OpenProcessToken(GetCurrentProcess(), TOKEN_QUERY, &hToken);
	BOOL isMember = FALSE;
	CheckTokenMembership(hToken, pAdminSid, &isMember);
	if(!isMember)
	{
		CheckTokenMembership(hToken, pAcctOpsSid, &isMember);
	}
	CloseHandle(hToken);
	return isMember;
}

HANDLE LogonUser(const std::wstring& userName, std::wstring password)
{
	ASSERT(!password.empty());
	WCHAR* pPassword = &password[0];
	UINT passByteLen = static_cast<UINT>(password.length() * sizeof(WCHAR));
	HANDLE hToken = NULL;
	RtlDecryptMemory(pPassword, passByteLen, 0);
	BOOL loggedOn = LogonUser(
		userName.c_str(),
		L".",
		password.c_str(),
		LOGON32_LOGON_INTERACTIVE,
		LOGON32_PROVIDER_DEFAULT,
		&hToken
	);
	SecureZeroMemory(pPassword, passByteLen);
	if(!loggedOn)
	{
		DBG_TEXT("Logon user failure - %lu", GetLastError());
	}
	return hToken;
}

PVOID CreateUserEnv(HANDLE hToken)
{
	PVOID pEnv = NULL;
	if(!CreateEnvironmentBlock(&pEnv, hToken, TRUE))
	{
		DBG_TEXT("Couldn't create environment block - %lu", GetLastError());
	}
	return pEnv;
}

typedef BOOL (WINAPI*pfnCreateWithToken)(
	HANDLE hToken,
	DWORD dwLogonFlags,
	LPCWSTR lpApplicationName,
	LPWSTR lpCommandLine,
	DWORD dwCreationFlags,
	LPVOID lpEnvironment,
	LPCWSTR lpCurrentDirectory,
	LPSTARTUPINFOW lpStartupInfo,
	LPPROCESS_INFORMATION lpProcessInfo
);

BOOL RunProcessAsUser(std::wstring& cmdLine, const CmdLineArgs& args, PROCESS_INFORMATION& pi)
{
	WinType<> token(LogonUser(args.userName, args.password), &CloseHandle);
	if(!*token)
	{
		return FALSE;
	}
	WinType<PVOID> pEnvBlock(CreateUserEnv(*token), &DestroyEnvironmentBlock);
	if(!*pEnvBlock)
	{
		return FALSE;
	}
	STARTUPINFO si = {sizeof(si), 0};
	si.dwFlags = STARTF_USESHOWWINDOW;
	si.wShowWindow = SW_SHOW;
	// the initial dir is in the context of the target process
	// so in essence, C:\ (or D:\, E:\ etc) is the chroot base dir
	WCHAR rootDir[4] = {FindARootedDriveLetter(args.drivesToKeep), L':', L'\\', 0};
	BOOL createRet = FALSE;
	HMODULE hMod = GetModuleHandle(L"advapi32.dll");
	ASSERT(hMod);
	pfnCreateWithToken createProcWithToken = reinterpret_cast<pfnCreateWithToken>(GetProcAddress(hMod, "CreateProcessWithTokenW"));
	if(createProcWithToken)
	{
		createRet = createProcWithToken(
			*token,
			LOGON_WITH_PROFILE,
			NULL,
			&cmdLine[0],
			CREATE_SUSPENDED | CREATE_UNICODE_ENVIRONMENT,
			*pEnvBlock,
			rootDir,
			&si,
			&pi
		);
	}
	else
	{
		std::wstring unencryptedPass = args.password;
		PWSTR pPass = &unencryptedPass[0];
		UINT passByteLen = static_cast<UINT>(unencryptedPass.length() * sizeof(WCHAR));
		RtlDecryptMemory(pPass, passByteLen, 0);
		createRet = CreateProcessWithLogonW(
			args.userName.c_str(), 
			L".", // current computer
			pPass,
			LOGON_WITH_PROFILE,
			NULL,
			&cmdLine[0],
			CREATE_SUSPENDED | CREATE_UNICODE_ENVIRONMENT,
			*pEnvBlock,
			rootDir,
			&si,
			&pi
		);
		// securely delete the password
		SecureZeroMemory(pPass, passByteLen);
	}
	return createRet;
}

void DeleteUser(const std::wstring& user)
{
	DWORD err = NetUserDel(NULL, user.c_str());
	if(err)
	{
		std::ostringstream stream;
		stream << "Couldn't delete user \'" << std::string(user.begin(), user.end()) << '\'';
		DisplayError(stream.str().c_str(), err);
	}
}

BOOL DoesUserExist(LPCWSTR pUserName)
{
	PBYTE pInfo = NULL;
	DWORD err = NetUserGetInfo(NULL, pUserName, 1, &pInfo);
	if(err != NERR_Success)
	{
		SetLastError(err);
	}
	else NetApiBufferFree(pInfo);
	return err == NERR_Success;
}